@@ -9,10 +9,11 @@
 # obtain one at https://mozilla.org/MPL/2.0/.
 
 import runpy
+import sys
 
 import pytest
 
-from hypothesis.internal import reflection
+from hypothesis.internal import lambda_sources
 from hypothesis.internal.conjecture.utils import identity
 from hypothesis.internal.reflection import get_pretty_function_description
 
@@ -24,9 +25,10 @@ def clear_lambda_caches(request, monkeypatch):
     # Run all tests in this file twice, once using cache and once forcing
     # from-scratch generation
     if request.param:
-        monkeypatch.setattr(reflection, "LAMBDA_DESCRIPTION_CACHE", {})
-        monkeypatch.setattr(reflection, "LAMBDA_DIGEST_DESCRIPTION_CACHE", {})
-        monkeypatch.setattr(reflection, "AST_LAMBDAS_CACHE", {})
+        monkeypatch.setattr(lambda_sources, "LAMBDA_DESCRIPTION_CACHE", {})
+        monkeypatch.setattr(lambda_sources, "LAMBDA_DIGEST_DESCRIPTION_CACHE", {})
+        monkeypatch.setattr(lambda_sources, "AST_LAMBDAS_CACHE", {})
+    return request.param
 
 
 def test_bracket_whitespace_is_stripped():
@@ -41,6 +43,9 @@ def test_can_have_unicode_in_lambda_sources():
     assert get_pretty_function_description(t) == "lambda x: 'é' not in x"
 
 
+@pytest.mark.skipif(
+    sys.version_info < (3, 11), reason="nested lambdas fail compile-test"
+)
 def test_can_get_descriptions_of_nested_lambdas_with_different_names():
     # fmt: off
     ordered_pair = (
@@ -64,6 +69,9 @@ def test_does_not_error_on_unparsable_source():
     assert get_pretty_function_description(t) == "lambda x: x"
 
 
+@pytest.mark.skipif(
+    sys.version_info < (3, 11), reason="nested lambdas fail compile-test"
+)
 def test_separate_line_map_filter():
     # this isn't intentionally testing nested lambdas, but hey, it's a nice bonus.
     # fmt: off
@@ -275,3 +283,79 @@ def test_modifying_lambda_source_code_returns_unknown(tmp_path):
         get_pretty_function_description(module_globals["test_lambda"])
         == "lambda x: <unknown>"
     )
+
+
+@skipif_threading  # concurrent writes to the same file
+def test_adding_other_lambda_does_not_confuse(tmp_path):
+    test_module = tmp_path / "test_module.py"
+    test_module.write_text(
+        "# line one\ntest_lambda = lambda x: x * 2", encoding="utf-8"
+    )
+    module_globals = runpy.run_path(str(test_module))
+
+    test_module.write_text(
+        "# line one\nlambda x: x\n\n\ntest_lambda = lambda x: x * 2", encoding="utf-8"
+    )
+    f1 = module_globals["test_lambda"]
+    assert get_pretty_function_description(f1) == "lambda x: x * 2"
+
+
+@skipif_threading  # concurrent writes to the same file
+def test_changing_lambda_confuses(tmp_path, allow_unknown_lambdas, clear_lambda_caches):
+    if not clear_lambda_caches:
+        pytest.skip("requires clean cache")
+    test_module = tmp_path / "test_module.py"
+    test_module.write_text("test_lambda = lambda x: x * 2", encoding="utf-8")
+    module_globals = runpy.run_path(str(test_module))
+
+    test_module.write_text("lambda x: x", encoding="utf-8")
+    f1 = module_globals["test_lambda"]
+    assert get_pretty_function_description(f1) == "lambda x: <unknown>"
+
+
+@skipif_threading  # concurrent writes to the same file
+@pytest.mark.skipif(sys.version_info[:2] < (3, 11), reason="not checked before 3.11")
+def test_that_test_harness_raises_on_unknown_lambda(tmp_path):
+    test_module = tmp_path / "test_module.py"
+    test_module.write_text("test_lambda = lambda x: x * 2", encoding="utf-8")
+    module_globals = runpy.run_path(str(test_module))
+
+    test_module.write_text("lambda x: x", encoding="utf-8")
+    f1 = module_globals["test_lambda"]
+    with pytest.raises(AssertionError):
+        assert get_pretty_function_description(f1) == "lambda x: <unknown>"
+
+
+@skipif_threading  # concurrent writes to the same file
+def test_source_with_syntax_error(tmp_path, allow_unknown_lambdas, clear_lambda_caches):
+    if not clear_lambda_caches:
+        pytest.skip("requires clean cache")
+    test_module = tmp_path / "test_module.py"
+    test_module.write_text("test_lambda = lambda x: x * 2", encoding="utf-8")
+    module_globals = runpy.run_path(str(test_module))
+    f1 = module_globals["test_lambda"]
+
+    test_module.write_text("line 1", encoding="utf-8")
+    assert get_pretty_function_description(f1) == "lambda x: <unknown>"
+
+
+@skipif_threading  # concurrent writes to the same file
+def test_unknown_is_not_stuck(tmp_path, allow_unknown_lambdas, clear_lambda_caches):
+    if not clear_lambda_caches:
+        pytest.skip("requires clean cache")
+    test_module = tmp_path / "test_module.py"
+    test_module.write_text("test_lambda = lambda x: x * 2", encoding="utf-8")
+    module_globals = runpy.run_path(str(test_module))
+    f1 = module_globals["test_lambda"]
+
+    test_module.write_text("line 1", encoding="utf-8")
+    assert get_pretty_function_description(f1) == "lambda x: <unknown>"
+
+    test_module2 = tmp_path / "test_module2.py"
+    test_module2.write_text("test_lambda = lambda x: x * 2", encoding="utf-8")
+    module_globals2 = runpy.run_path(str(test_module2))
+    f2 = module_globals2["test_lambda"]
+
+    # f2 matches f1 in the digest cache, ensure that it gets a proper description
+    assert get_pretty_function_description(f1) == "lambda x: <unknown>"
+    assert get_pretty_function_description(f2) == "lambda x: x * 2"
